
import asyncio
import logging
from typing import List, Optional

import a_sync
from multicall import Call

from y.classes.common import ERC20, WeiBalance
from y.contracts import has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice

logger = logging.getLogger(__name__)


@a_sync.a_sync(default='sync', cache_type='memory')
async def is_token_set(token: AnyAddressType) -> bool:
    return any(await asyncio.gather(
        has_methods(token, ("getComponents()(address[])", "naturalUnit()(uint)"), sync=False),
        has_methods(token, ("getComponents()(address[])", "getModules()(address[])", "getPositions()(address[])"), sync=False),
    ))

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await TokenSet(token, asynchronous=True).get_price(block=block)


class TokenSet(ERC20):
    @a_sync.a_sync(ram_cache_maxsize=100)
    async def components(self, block: Optional[Block] = None) -> List[ERC20]:
        components = await self.contract.getComponents.coroutine(block_identifier = block)
        return [ERC20(component, asynchronous=self.asynchronous) for component in components]
    
    async def balances(self, block: Optional[Block] = None) -> List[WeiBalance]:
        if hasattr(self.contract, 'getUnits'):
            balances = await self.contract.getUnits.coroutine(block_identifier = block)
            logger.debug("getUnits: %s", balances)
        elif hasattr(self.contract, 'getTotalComponentRealUnits'):
            balances = await asyncio.gather(*[
                Call(self.address, ["getTotalComponentRealUnits(address)(int256)", component.address], block_id=block).coroutine()
                for component in await self.components(block, sync=False)
            ])
            logger.debug("getTotalComponentRealUnits: %s", balances)
        balances = [WeiBalance(balance, component, block=block) for component, balance in zip(await self.components(block=block, sync=False), balances)]
        return balances
    
    async def get_price(self, block: Optional[Block] = None) -> UsdPrice:
        total_supply = await self.total_supply_readable(block=block, sync=False)
        if total_supply == 0:
            logger.debug("total supply is 0, forcing price to $0")
            return UsdPrice(0)
        if hasattr(self.contract, "getUnits"):
            balances = await self.balances(block=block, sync=False)
            values = await asyncio.gather(*[balance.__value_usd__(sync=False) for balance in balances])
            logger.debug("balances: %s  values: %s", balances, values)
            tvl = sum(values)
            price = UsdPrice(tvl / total_supply)
            logger.debug("total supply: %s  tvl: %s  price: %s", total_supply, tvl, price)
            return price
        elif hasattr(self.contract, "getTotalComponentRealUnits"):
            balances_per_token = await self.balances(block=block, sync=False)
            values_per_token = await asyncio.gather(*[balance.__value_usd__(sync=False) for balance in balances_per_token])
            price = UsdPrice(sum(values_per_token))
            logger.debug("balances per token: %s  values per token: %s  price: %s", balances_per_token, values_per_token, price)
            return price
        raise NotImplementedError



import logging

import a_sync
from async_lru import alru_cache
from brownie import web3
from cachetools.func import lru_cache
from web3._utils.rpc_abi import RPC

from y.utils.dank_mids import dank_w3
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)

@lru_cache(1)
@yLazyLogger(logger)
def get_ethereum_client() -> str:
    return _get_ethereum_client(web3.clientVersion)

@alru_cache(maxsize=1)
async def get_ethereum_client_async() -> str:
    return _get_ethereum_client(await dank_w3.manager.coro_request(RPC.web3_clientVersion, []))

@yLazyLogger(logger)
def _get_ethereum_client(client: str) -> str:
    if client.startswith('TurboGeth'):
        return 'tg'
    if client.lower().startswith('erigon'):
        return 'erigon'
    if client.lower().startswith('geth'):
        return 'geth'
    logger.debug("client: %s", client)




from datetime import datetime

from pony.orm import (Database, Optional, PrimaryKey, Required, Set,
                      composite_key)

db = Database()


#E = TypeVar("E", bound=db.Entity)

class _AsyncEntityMixin:
    pass
    '''
    @classmethod
    @a_sync('async')
    def get(cls: E, *args, **kwargs) -> E:
        return super(db.Entity).get(*args, **kwargs)
    @classmethod
    @a_sync('async')
    def select(cls: E, *args, **kwargs) -> E:
        return super(db.Entity).select(*args, **kwargs)'''
    
class Chain(db.Entity, _AsyncEntityMixin):
    _pk = PrimaryKey(int, auto=True)
    id = Required(int, unique=True, lazy=True)

    blocks = Set("Block")
    addresses = Set("Address")

class Block(db.Entity, _AsyncEntityMixin):
    _pk = PrimaryKey(int, auto=True)
    chain = Required(Chain, reverse="blocks")
    number = Required(int, lazy=True)
    hash = Optional(int, lazy=True)
    timestamp = Optional(datetime, lazy=True)
    
    composite_key(chain, number)
    composite_key(chain, hash)

    contracts_deployed = Set("Contract", reverse="deploy_block")

class Address(db.Entity, _AsyncEntityMixin):
    _pk = PrimaryKey(int, auto=True)
    chain = Required(Chain, lazy=True, reverse="addresses")
    address = Required(str, lazy=True)
    notes = Optional(str, lazy=True)
    
    composite_key(chain, address)

    contracts_deployed = Set("Contract", reverse="deployer")

class Contract(Address):
    deployer = Optional(Address, reverse='contracts_deployed', lazy=True)
    deploy_block = Optional(Block, reverse='contracts_deployed', lazy=True)

class Token(Contract):
    symbol = Optional(str, lazy=True)
    name = Optional(str, lazy=True)
    decimals = Optional(int, lazy=True)
    bucket = Optional(str, lazy=True)




import asyncio
from contextlib import suppress
from functools import cached_property
from typing import Any, Optional, Tuple, Union

import a_sync
from brownie.convert.datatypes import HexString
from brownie.exceptions import ContractNotFound

from y import convert
from y.classes.singleton import ChecksumASyncSingletonMeta
from y.constants import EEE_ADDRESS
from y.contracts import Contract, build_name, has_method, probe
from y.datatypes import AnyAddressType, Block, Pool, UsdPrice
from y.erc20 import decimals, totalSupply
from y.exceptions import (ContractNotVerified, MessedUpBrownieContract,
                          NonStandardERC20)
from y.networks import Network
from y.utils import logging, raw_calls


def hex_to_string(h: HexString) -> str:
    '''returns a string from a HexString'''
    h = h.hex().rstrip("0")
    if len(h) % 2 != 0:
        h += "0"
    return bytes.fromhex(h).decode("utf-8")

class ContractBase(a_sync.ASyncGenericBase, metaclass=ChecksumASyncSingletonMeta):
    def __init__(self, address: AnyAddressType, asynchronous: bool = False) -> None:
        self.address = convert.to_address(address)
        self.asynchronous = asynchronous
        super().__init__()
    
    def __str__(self) -> str:
        return f'{self.address}'

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} '{self.address}'"
    
    def __eq__(self, __o: object) -> bool:
        try:
            return convert.to_address(__o) == self.address
        except Exception:
            return False
    
    def __hash__(self) -> int:
        return hash(self.address)
    
    @property
    def contract(self) -> Contract:
        return Contract(self.address)
    
    @cached_property
    def _is_cached(self) -> bool:
        try:
            self.contract
            return True
        except (ContractNotVerified):
            return False
        except (ContractNotFound, MessedUpBrownieContract):
            return None

    @a_sync.aka.cached_property
    async def build_name(self) -> str:
        return await build_name(self.address, sync=False)

    async def has_method(self, method: str, return_response: bool = False) -> Union[bool,Any]:
        return await has_method(self.address, method, return_response=return_response, sync=False)


class ERC20(ContractBase):
    def __repr__(self) -> str:
        cls = self.__class__.__name__
        with suppress(AttributeError):
            if ERC20.symbol.has_cache_value(self):
                symbol = ERC20.symbol.get_cache_value(self)
                return f"<{cls} {symbol} '{self.address}'>"
            elif not asyncio.get_event_loop().is_running() and not self.asynchronous:
                return f"<{cls} {self.symbol} '{self.address}'>"
        return super().__repr__()
    
    @a_sync.aka.cached_property
    async def symbol(self) -> str:
        if self.address == EEE_ADDRESS:
            return "ETH"
        
        import y._db.utils as db

        symbol = await db._get_token_symbol(self.address)
        if symbol:
            return symbol
        
        symbol = await probe(self.address, ["symbol()(string)", "SYMBOL()(string)", "getSymbol()(string)"])
        if symbol is None:
            # Sometimes the above will fail if the symbol method returns bytes32, as with MKR. Let's try this.
            symbol = await probe(self.address, ["symbol()(bytes32)"])
            if symbol:
                symbol = hex_to_string(symbol)
        
        if symbol:
            await db._set_token_symbol(self.address, symbol)
            return symbol
        
        # we've failed to fetch
        self.__raise_exception('symbol')
    
    @a_sync.aka.property
    async def name(self) -> str:
        if self.address == EEE_ADDRESS:
            return "Ethereum"
        
        import y._db.utils as db

        name = await db._get_token_name(self.address)
        if name:
            return name
        
        name = await probe(self.address, ["name()(string)", "NAME()(string)", "getName()(string)"])
        if name is None:
            # Sometimes the above will fail if the name method returns bytes32, as with MKR. Let's try this.
            name = await probe(self.address, ["name()(bytes32)"])
            if name:
                name = hex_to_string(name)
        
        if name:
            await db._set_token_name(self.address, name)
            return name
                
        # we've failed to fetch
        self.__raise_exception('name')
    
    @a_sync.aka.cached_property
    async def decimals(self) -> int:
        if self.address == EEE_ADDRESS:
            return 18
        import y._db.utils as db
        return await db.get_token_decimals(self.address)

    @a_sync.a_sync # Override the leading underscore so a_sync lib doesn't bypass this fn
    async def _decimals(self, block: Optional[Block] = None) -> int:
        '''used to fetch decimals at specific block'''
        if self.address == EEE_ADDRESS:
            return 18
        return await decimals(self.address, block=block, sync=False)
    
    @a_sync.aka.cached_property
    async def scale(self) -> int:
        return 10 ** await self.__decimals__(asynchronous=True)
    
    @a_sync.a_sync # Override the leading underscore so a_sync lib doesn't bypass this fn
    async def _scale(self, block: Optional[Block] = None) -> int:
        return 10 ** await self._decimals(block, sync=False)

    async def total_supply(self, block: Optional[Block] = None) -> int:
        return await totalSupply(self.address, block=block, sync=False)

    async def total_supply_readable(self, block: Optional[Block] = None) -> float:
        total_supply, scale = await asyncio.gather(
            self.total_supply(block=block, sync=False),
            self.__scale__(sync=False)
        )
        return total_supply / scale
    
    async def balance_of(self, address: AnyAddressType, block: Optional[Block] = None) -> int:
        return await raw_calls.balanceOf(self.address, address, block=block, sync=False)
    
    async def balance_of_readable(self, address: AnyAddressType, block: Optional[Block] = None) -> float:
        balance, scale = await asyncio.gather(self.balance_of(address, block=block, asynchronous=True), self.__scale__(asynchronous=True))
        return balance / scale

    async def price(
        self, 
        block: Optional[Block] = None, 
        return_None_on_failure: bool = False, 
        ignore_pools: Tuple[Pool, ...] = (),
    ) -> Optional[UsdPrice]:
        from y.prices.magic import get_price
        return await get_price(
            self.address, 
            block=block, 
            fail_to_None=return_None_on_failure,
            ignore_pools=ignore_pools,
            sync=False,
        )
    
    def __raise_exception(self, fn_name: str):
        raise NonStandardERC20(f'''
            Unable to fetch `{fn_name}` for {self.address} on {Network.printable()}
            If the contract is verified, please check to see if it has a strangely named
            `{fn_name}` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
            with the contract address and correct method name so we can keep things going smoothly :)''')


class WeiBalance(a_sync.ASyncGenericBase):
    def __init__(
        self, 
        balance: int,
        token: AnyAddressType,
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
        asynchronous: bool = False,
        ) -> None:

        self.asynchronous = asynchronous
        self.balance = balance
        self.token = ERC20(str(token), asynchronous=self.asynchronous)
        self.block = block
        super().__init__()
        self._logger = logging.get_price_logger(token, block, self.__class__.__name__)
        self._ignore_pools = ignore_pools

    def __str__(self) -> str:
        return str(self.balance)
    
    def __repr__(self) -> str:
        return f"<WeiBalance token={self.token} balance={self.balance} block={self.block}>"

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, int):
            return __o == self.balance
        elif isinstance(__o, WeiBalance):
            return self.balance == __o.balance and self.token == __o.token
        return False
    
    def __lt__(self, __o: object) -> bool:
        if isinstance(__o, int):
            return __o < self.balance
        elif isinstance(__o, WeiBalance):
            if self.token != __o.token:
                raise TypeError(f"'<' only supported between {self.__class__.__name__} instances when denominated in the same token.")
            return self.balance < __o.balance
        raise TypeError(f"'<' not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'")

    def __ge__(self, __o: object) -> bool:
        if __o < self:
            return True
        elif type(__o) is type(self):
            return self == __o
        raise TypeError(f"'>=' not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'")
    
    @a_sync.aka.property
    async def readable(self) -> float:
        if self.balance == 0:
            return 0
        scale = await self.token.__scale__(sync=False)
        readable = self.balance / scale
        self._logger.debug("balance: %s  decimals: %s  readable: %s", self.balance, str(scale).count("0"), readable)
        return readable
    
    @a_sync.aka.cached_property
    async def value_usd(self) -> float:
        if self.balance == 0:
            return 0
        balance, price = await asyncio.gather(
            self.__readable__(sync=False),
            self.token.price(block=self.block, ignore_pools=self._ignore_pools, sync=False),
        )
        value = balance * price
        self._logger.debug("balance: %s  price: %s  value: %s", balance, price, value)
        return value



import asyncio
import logging
from typing import Optional, Tuple

import a_sync
from brownie import chain, convert
from brownie.exceptions import VirtualMachineError
from multicall import Call

from y.classes.common import ERC20, ContractBase
from y.constants import EEE_ADDRESS
from y.contracts import Contract, has_methods
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice
from y.exceptions import call_reverted
from y.networks import Network
from y.utils.logging import _gh_issue_request
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)

TROLLERS = {
    Network.Mainnet: {
        "comp":             "0x3d9819210A31b4961b30EF54bE2aeD79B9c9Cd3B",
        "cream":            "0x3d5BC3c8d13dcB8bF317092d84783c2697AE9258",
        "ironbank":         "0xAB1c342C7bf5Ec5F02ADEA1c2270670bCa144CbB",
        "inverse":          "0x4dCf7407AE5C07f8681e1659f626E114A7667339",
        "unfederalreserve": "0x3105D328c66d8d55092358cF595d54608178E9B5",
        "flux":             "0x95Af143a021DF745bc78e845b54591C53a8B3A51",
    },
    Network.BinanceSmartChain: {
        "venus":            "0xfD36E2c2a6789Db23113685031d7F16329158384",
    },
    Network.Polygon: {
        "easyfi":           "0xcb3fA413B23b12E402Cfcd8FA120f983FB70d8E8",
        "apple":            "0x46220a07F071D1a821D68fA7C769BCcdA3C65430",
        "chumhum":          "0x1D43f6DA91e9EF6614dCe95bCef43E4d7b2bcFB5",
        "cream":            "0x20CA53E2395FA571798623F1cFBD11Fe2C114c24",
    },
    Network.Fantom: {
        "cream":            "0x4250A6D3BD57455d7C6821eECb6206F507576cD2",
        "scream":           "0x260E596DAbE3AFc463e75B6CC05d8c46aCAcFB09",
        "ola":              "0xD65eB596cFb5DE402678a12df651E0e588Dc3A81",
    },
    Network.Avalanche: {
        "vee":              "0xA67DFeD73025b0d61F2515c531dd8D25D4Cfd0Db",
        "vee2":             "0x43AAd7d8Bc661dfA70120865239529ED92Faa054",
        "vee3":             "0xeEf69Cab52480D2BD2D4A3f3E8F5CcfF2923f6eF",
        "cream":            "0x2eE80614Ccbc5e28654324a66A396458Fa5cD7Cc",
    },
    Network.Arbitrum: {
        "cream":            "0xbadaC56c9aca307079e8B8FC699987AAc89813ee",
        "neku":             "0xD5B649c7d27C13a2b80425daEe8Cb6023015Dc6B",
        "channels":         "0x3C13b172bf8BE5b873EB38553feC50F78c826284",
        "hund":             "0x0F390559F258eB8591C8e31Cf0905E97cf36ACE2",
    },
    Network.Optimism: {
        "ironbank":         "0xE0B57FEEd45e7D908f2d0DaCd26F113Cf26715BF",
    }
}.get(chain.id, {})


class CToken(ERC20):
    def __init__(self, address: AnyAddressType, comptroller: Optional["Comptroller"] = None, asynchronous: bool = False) -> None:
        self.troller = comptroller
        super().__init__(address, asynchronous=asynchronous)
    
    async def get_price(self, block: Optional[Block] = None) -> UsdPrice:
        if self.troller:
            # We can use the protocol's oracle which will be quick (if it works)
            underlying_per_ctoken, underlying_price = await asyncio.gather(
                self.underlying_per_ctoken(block=block, asynchronous=True),
                self.get_underlying_price(block=block, asynchronous=True),
            )
            if underlying_price:
                return UsdPrice(underlying_per_ctoken * underlying_price)
            
        # Or we can just price the underlying token ourselves
        underlying = await self.__underlying__(asynchronous=True)
        underlying_per_ctoken, underlying_price = await asyncio.gather(
            self.underlying_per_ctoken(block=block, asynchronous=True),
            underlying.price(block=block, asynchronous=True)
        )
        return UsdPrice(underlying_per_ctoken * underlying_price)
    
    @a_sync.aka.cached_property
    async def underlying(self) -> ERC20:
        underlying = await self.has_method('underlying()(address)', return_response=True, sync=False)
        # this will run for gas coin markets like cETH, crETH
        if not underlying:
            underlying = EEE_ADDRESS
        return ERC20(underlying, asynchronous=self.asynchronous)
    
    async def underlying_per_ctoken(self, block: Optional[Block] = None) -> float:
        exchange_rate, decimals, underlying = await asyncio.gather(
            self.exchange_rate(block=block, sync=False),
            self.__decimals__(sync=False),
            self.__underlying__(sync=False),
        )
        return exchange_rate * 10 ** (decimals - await underlying.__decimals__(asynchronous=True))

    #yLazyLogger(logger)
    async def exchange_rate(self, block: Optional[Block] = None) -> float:
        method = 'exchangeRateCurrent()(uint)'
        call = Call(self.address, [method], block_id=block)

        try:
            exchange_rate = await call.coroutine()
        except Exception as e:
            if not call_reverted(e):
                raise e
            exchange_rate = None

        if exchange_rate is None:
            # NOTE: Sometimes this works, not sure why
            try:
                exchange_rate = self.contract.exchangeRateCurrent.call(block_identifier=block)
            except Exception as e:
                if 'borrow rate is absurdly high' not in str(e):
                    raise
                exchange_rate = 0
        
        return exchange_rate / 1e18
    
    async def get_underlying_price(self, block: Optional[Block] = None) -> Optional[float]:
        # always query the oracle in case it was changed
        oracle, underlying = await asyncio.gather(
            self.troller.oracle(block, asynchronous=True),
            self.__underlying__(asynchronous=True),
        )
        price, underlying_decimals = await asyncio.gather(
            oracle.getUnderlyingPrice.coroutine(self.address, block_identifier=block),
            underlying.__decimals__(asynchronous=True),
            return_exceptions=True,
        )
        if isinstance(price, Exception):
            # TODO debug why this occurs and refactor. only found on arbitrum cream
            try:
                price = oracle.getUnderlyingPrice(self.address, block_identifier=block)
            except VirtualMachineError as e:
                if str(e) in {
                    "revert: grace period not over",
                    "revert: Chainlink feeds are not being updated",
                }:
                    return None
                raise e
        price /= 10 ** (36 - underlying_decimals)
        return price
    

class Comptroller(ContractBase):
    def __init__(self, address: Optional[AnyAddressType] = None, key: Optional[str] = None, asynchronous: bool = False) -> None:
        assert address or key,          'Must provide either an address or a key'
        assert not (address and key),   'Must provide either an address or a key, not both'

        if key: address = TROLLERS[key]
        else: key = [key for key in TROLLERS if address == TROLLERS[key]][0]

        self.address = convert.to_address(address)
        self.key = key
        self.asynchronous = asynchronous
    
    def __repr__(self) -> str:
        return f"<Comptroller {self.key} '{self.address}'>"

    #yLazyLogger(logger)
    def __contains__(self, token_address: AnyAddressType) -> bool:
        if self.asynchronous:
            raise RuntimeError(f"'self.asynchronous' must be False to use Comptroller.__contains__")
        return token_address in self.markets

    @a_sync.aka.cached_property
    async def markets(self) -> Tuple[CToken]:
        response = await self.has_method("getAllMarkets()(address[])", return_response=True, sync=False)
        if not response:
            logger.warning(f'had trouble loading markets for {self.__repr__()}')
            response = set()
        markets = tuple(CToken(market, comptroller=self, asynchronous=self.asynchronous) for market in response)
        logger.info("loaded %s markets for %s", len(markets), self.__repr__())
        return markets
    
    async def oracle(self, block: Optional[Block] = None) -> Contract:
        try:
            oracle = await self.contract.oracle.coroutine(block_identifier=block)
        except Exception as e:
            # TODO debug why this occurs and refactor. only found on arbitrum cream
            if not call_reverted(e):
                raise
            oracle = self.contract.oracle(block_identifier=block)
        return await Contract.coroutine(oracle)


class Compound(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
        self.trollers = {
            protocol: Comptroller(troller, asynchronous=self.asynchronous)
            for protocol, troller
            in TROLLERS.items()
        }
    
    def __contains__(self, token_address: AddressOrContract) -> bool:
        if self.asynchronous:
            raise RuntimeError("'self.asynchronous' must be False and the event loop must not be running")
        return self.is_compound_market(token_address)
    
    async def get_troller(self, token_address: AddressOrContract) -> Optional[Comptroller]:
        trollers = self.trollers.values()
        all_markets = await asyncio.gather(*[troller.__markets__(sync=False) for troller in trollers])
        for troller, markets in zip(trollers, all_markets):
            if token_address in markets:
                return troller
            
    async def is_compound_market(self, token_address: AddressOrContract) -> bool:
        if await self.get_troller(token_address, sync=False):
            return True

        # NOTE: Workaround for pools that have since been revoked
        result = await has_methods(token_address, ('isCToken()(bool)','comptroller()(address)','underlying()(address)'), sync=False)
        if result is True:
            await self.__notify_if_unknown_comptroller(token_address)
        return result
    
    async def get_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        troller = await self.get_troller(token_address)
        return await CToken(token_address, comptroller=troller, asynchronous=True).get_price(block=block)

    async def __notify_if_unknown_comptroller(self, token_address: AddressOrContract) -> None:
        comptroller = await raw_call(token_address,'comptroller()',output='address', sync=False)
        if comptroller not in self.trollers.values():
            _gh_issue_request(f'Comptroller {comptroller} is unknown to ypricemagic.', logger)


compound = Compound(asynchronous=True)




import logging
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from typing import Optional

from a_sync import a_sync
from brownie import chain, convert
from pony.orm import (BindingError, TransactionError,
                      TransactionIntegrityError, commit, db_session)

from y._db.config import connection_settings
from y._db.entities import Address, Chain, Token, db
from y.constants import EEE_ADDRESS
from y.erc20 import decimals

logger = logging.getLogger(__name__)

try:
    db.bind(**connection_settings, create_db=True)
    db.generate_mapping(create_tables=True)
except TransactionError as e:
    if str(e) != "@db_session-decorated create_tables() function with `ddl` option cannot be called inside of another db_session":
        raise e
except BindingError as e:
    if not str(e).startswith('Database object was already bound to'):
        raise e


async def get_token_decimals(address: str) -> int:
    d = await _get_token_decimals(address)
    if d is None:
        d = await decimals(address, sync=False)
        await _set_token_decimals(address, d)
    return d


executor = ThreadPoolExecutor(16)

@a_sync(default='async', executor=executor)
@db_session
def get_chain() -> Chain:
    if c:=Chain.get(id=chain.id):
        return c
    with suppress(TransactionIntegrityError):
        Chain(id=chain.id)
        commit()
        logger.debug('chain %s added to ydb', chain.id)
    return Chain.get(id=chain.id)

@a_sync(default='async', executor=executor)
@db_session
def get_token(address: str) -> Token:
    address = convert.to_address(address)
    if address == EEE_ADDRESS:
        raise ValueError(f"cannot create token entity for {EEE_ADDRESS}")
    while True:
        if entity := Address.get(chain=get_chain(sync=True), address=address):
            if isinstance(entity, Token):
                return entity
            entity.delete()
            commit()
        with suppress(TransactionIntegrityError):
            Token(chain=get_chain(sync=True), address=address)
            commit()
            logger.debug('token %s added to ydb', address)
        if token := Token.get(chain=get_chain(sync=True), address=address):
            return token
    return token

@a_sync(default='async', executor=executor)
@db_session
def _get_token_symbol(address: str) -> Optional[str]:
    return get_token(address, sync=True).symbol

@a_sync(default='async', executor=executor)
@db_session
def _set_token_symbol(address: str, symbol: str) -> None:
    get_token(address, sync=True).symbol = symbol

@a_sync(default='async', executor=executor)
@db_session
def _get_token_name(address: str) -> Optional[str]:
    return get_token(address, sync=True).name

@a_sync(default='async', executor=executor)
@db_session
def _set_token_name(address: str, name: str) -> None:
    get_token(address, sync=True).name = name

@a_sync(default='async', executor=executor)
@db_session
def _get_token_decimals(address: str) -> Optional[int]:
    return get_token(address, sync=True).decimals

@a_sync(default='async', executor=executor)
@db_session
def _set_token_decimals(address: str, decimals: int) -> None:
    get_token(address, sync=True).decimals = decimals

@a_sync(default='async', executor=executor)
@db_session
def _get_token_bucket(address: str) -> Optional[str]:
    if address == EEE_ADDRESS:
        return
    return get_token(address, sync=True).bucket

@a_sync(default='async', executor=executor)
@db_session
def _set_token_bucket(address: str, bucket: str) -> None:
    if address == EEE_ADDRESS:
        return
    get_token(address, sync=True).bucket = bucket



import asyncio
import logging
from typing import Awaitable, Dict, List, NewType, Optional, Tuple

import a_sync
from brownie import chain
from brownie.convert.datatypes import EthAddress
from hexbytes import HexBytes
from multicall import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y import constants, contracts
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.datatypes import Address, AnyAddressType, Block, UsdPrice, UsdValue
from y.networks import Network
from y.utils.events import decode_logs, get_logs_asap
from y.utils.logging import get_price_logger
from y.utils.raw_calls import raw_call

BALANCER_V2_VAULTS = {
    Network.Mainnet: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Fantom: [
        '0x20dd72Ed959b6147912C2e529F0a0C651c33c9ce',
    ],
    Network.Polygon: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Arbitrum: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Base: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
}.get(chain.id, [])


PoolId = NewType('PoolId', bytes)

class BalancerV2Vault(ContractBase):
    def __init__(self, address: AnyAddressType, asynchronous: bool = False) -> None:
        super().__init__(address, asynchronous=asynchronous)
        if not self._is_cached:
            # we need the contract cached so we can decode logs correctly
            self.contract
            
    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_pool_tokens(self, pool_id: int, block: Optional[Block] = None):
        return await self.contract.getPoolTokens.coroutine(pool_id, block_identifier = block)
    
    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def list_pools(self, block: Optional[Block] = None) -> Dict[HexBytes,EthAddress]:
        topics = ['0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e']
        events = decode_logs(await get_logs_asap(self.address, topics, to_block=block, sync=False))
        return {
            event['poolId'].hex(): event['poolAddress'] for event in events
            # NOTE: For some reason the Balancer fork on Fantom lists "0x3e522051A9B1958Aa1e828AC24Afba4a551DF37d"
            #       as a pool, but it is not a contract. This handler will prevent it and future cases from causing problems.
            if contracts.is_contract(event['poolAddress'])
        }
    
    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_pool_info(self, poolids: Tuple[HexBytes,...], block: Optional[Block] = None) -> List[Tuple]:
        return await asyncio.gather(*[
            self.contract.getPoolTokens.coroutine(poolId, block_identifier=block)
            for poolId in poolids
        ])
    
    async def deepest_pool_for(self, token_address: Address, block: Optional[Block] = None) -> Tuple[Optional[EthAddress],int]:
        # sourcery skip: simplify-len-comparison
        pools = await self.list_pools(block=block, sync=False)
        poolids = tuple(pools.keys())
        pools_info = await self.get_pool_info(poolids, block=block, sync=False)
        all_pools = await self.list_pools(block=block, sync=False)
        pools_info = {all_pools[poolid]: info for poolid, info in zip(poolids, pools_info) if str(info) != "((), (), 0)"}
        
        logger = get_price_logger(token_address, block, 'balancer.v2')
        deepest_pool = {'pool': None, 'balance': 0}
        for pool, info in pools_info.items():
            num_tokens = len(info[0])
            pool_balances = {info[0][i]: info[1][i] for i in range(num_tokens)}
            pool_balance = [balance for token, balance in pool_balances.items() if token == token_address]
            if len(pool_balance) == 0:
                continue
            assert len(pool_balance) == 1
            pool_balance = pool_balance[0]
            if pool_balance > deepest_pool['balance']:
                deepest_pool = {'pool': pool, 'balance': pool_balance}
        logger.debug("deepest pool %s", deepest_pool)
        return tuple(deepest_pool.values())


class BalancerV2Pool(ERC20):
    def __init__(self, pool_address: AnyAddressType, asynchronous: bool = False) -> None:
        super().__init__(pool_address, asynchronous=asynchronous)

    @a_sync.aka.cached_property
    async def id(self) -> PoolId:
        return PoolId(await Call(self.address, ['getPoolId()(bytes32)']).coroutine())
    
    @a_sync.aka.cached_property
    async def vault(self) -> BalancerV2Vault:
        vault = await raw_call(self.address,'getVault()',output='address', sync=False)
        return BalancerV2Vault(vault, asynchronous=True)
    
    async def get_pool_price(self, block: Optional[Block] = None) -> Awaitable[UsdPrice]:
        tvl, total_supply = await asyncio.gather(
            self.get_tvl(block=block, sync=False),
            self.total_supply_readable(block=block, sync=False),
        )
        return UsdPrice(tvl / total_supply)
        
    async def get_tvl(self, block: Optional[Block] = None) -> Awaitable[UsdValue]:
        balances = await self.get_balances(block=block, sync=False)
        return UsdValue(sum(await asyncio.gather(*[
            balance.__value_usd__(sync=False) for balance in balances.values()
            if balance.token.address != self.address  # NOTE: to prevent an infinite loop for tokens that include themselves in the pool (e.g. bb-a-USDC)
        ])))

    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_balances(self, block: Optional[Block] = None) -> Dict[ERC20, WeiBalance]:
        tokens = await self.tokens(block=block, sync=False)
        return dict(tokens.items())

    async def get_token_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        token_balances, weights = await asyncio.gather(
            self.get_balances(block=block, sync=False),
            self.weights(block=block, sync=False),
        )
        pool_token_info = list(zip(token_balances.keys(),token_balances.values(), weights))
        for pool_token, balance, weight in pool_token_info:
            if pool_token == token_address:
                token_balance, token_weight = balance, weight

        paired_token_balance = None
        for pool_token, balance, weight in pool_token_info:
            if pool_token in constants.STABLECOINS:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif pool_token == constants.WRAPPED_GAS_COIN:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif len(pool_token_info) == 2 and pool_token != token_address:
                paired_token_balance, paired_token_weight = balance, weight
                break
        
        if paired_token_balance is None:
            return None

        token_value_in_pool, token_balance_readable = await asyncio.gather(*[
            paired_token_balance.__value_usd__(sync=False),
            token_balance.__readable__(sync=False),
        ])
        token_value_in_pool /= paired_token_weight * token_weight
        return UsdPrice(token_value_in_pool / token_balance_readable) 

    # NOTE: We can't cache this as a cached property because some balancer pool tokens can change. Womp
    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def tokens(self, block: Optional[Block] = None) -> Dict[ERC20, WeiBalance]:
        vault, id = await asyncio.gather(self.__vault__(sync=False), self.__id__(sync=False))
        tokens, balances, lastChangedBlock = await vault.get_pool_tokens(id, block=block, sync=False)
        return {ERC20(token, asynchronous=self.asynchronous): WeiBalance(balance, token, block=block) for token, balance in zip(tokens, balances)}

    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def weights(self, block: Optional[Block] = None) -> List[int]:
        try:
            return await self.contract.getNormalizedWeights.coroutine(block_identifier = block)
        except (AttributeError,ValueError):
            tokens = await self.tokens(block=block, sync=False)
            return [1 for _ in tokens.keys()]


#yLazyLogger(logger)
@a_sync.a_sync(cache_type='memory')
async def _is_standard_pool(pool: EthAddress) -> bool:
    '''
    Returns `False` if `build_name(pool) in ['ConvergentCurvePool','MetaStablePool']`, else `True`
    '''
    
    # With `return_None_on_failure=True`, if `build_name(pool)` fails,
    # we can't know for sure that its a standard pool, but... it probably is.
    return await contracts.build_name(pool, return_None_on_failure=True, sync=False) not in ['ConvergentCurvePool','MetaStablePool']


class BalancerV2(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
        self.vaults = [BalancerV2Vault(vault, asynchronous=self.asynchronous) for vault in BALANCER_V2_VAULTS]
    
    def __str__(self) -> str:
        return "BalancerV2()"

    async def is_pool(self, token_address: AnyAddressType) -> bool:
        methods = ('getPoolId()(bytes32)','getPausedState()((bool,uint,uint))','getSwapFeePercentage()(uint)')
        return await contracts.has_methods(token_address, methods, sync=False)
    
    async def get_pool_price(self, pool_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        return await BalancerV2Pool(pool_address, asynchronous=True).get_pool_price(block=block)

    async def get_token_price(self, token_address: Address, block: Optional[Block] = None) -> UsdPrice:
        deepest_pool: Optional[BalancerV2Pool] = await self.deepest_pool_for(token_address, block=block, sync=False)
        if deepest_pool is None:
            return
        return await deepest_pool.get_token_price(token_address, block, sync=False)
    
    async def deepest_pool_for(self, token_address: Address, block: Optional[Block] = None) -> Optional[BalancerV2Pool]:
        deepest_pools = await asyncio.gather(*[vault.deepest_pool_for(token_address, block=block, sync=False) for vault in self.vaults])
        deepest_pools = {vault.address: deepest_pool for vault, deepest_pool in zip(self.vaults, deepest_pools) if deepest_pool is not None}
        deepest_pool_balance = max(pool_balance for pool_address, pool_balance in deepest_pools.values())
        for pool_address, pool_balance in deepest_pools.values():
            if pool_balance == deepest_pool_balance and pool_address:
                return BalancerV2Pool(pool_address, asynchronous=self.asynchronous)

balancer = BalancerV2(asynchronous=True)




from typed_envs import EnvVarFactory

_envs = EnvVarFactory("YPRICEMAGIC")

# TTL for various in-memory caches thruout the library
CACHE_TTL = _envs.create_env("CACHE_TTL", int, default=60*60, verbose=False)
CONTRACT_CACHE_TTL = _envs.create_env("CONTRACT_CACHE_TTL", int, default=int(CACHE_TTL), verbose=False)

DB_PROVIDER = _envs.create_env("DB_PROVIDER", str, default="sqlite", verbose=False)
DB_HOST = _envs.create_env("DB_HOST", str, default="", verbose=False)
DB_USER = _envs.create_env("CACHE_TTL", str, default="", verbose=False)
DB_PASSWORD = _envs.create_env("CACHE_TTL", str, default="", verbose=False)
DB_DATABASE = _envs.create_env("CACHE_TTL", str, default="ypricemagic", verbose=False)



import asyncio
import math
from itertools import cycle
from typing import List, Optional, Tuple

import a_sync
from brownie import chain
from eth_abi.packed import encode_abi_packed

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, ContractBase
from y.constants import usdc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import ContractNotVerified, TokenNotFound, UnsupportedNetwork
from y.interfaces.uniswap.quoterv3 import UNIV3_QUOTER_ABI
from y.networks import Network
from y.utils.events import decode_logs, get_logs_asap_generator
from y.utils.multicall import fetch_multicall

# https://github.com/Uniswap/uniswap-v3-periphery/blob/main/deploys.md
UNISWAP_V3_FACTORY = '0x1F98431c8aD98523631AE4a59f267346ea31F984'
UNISWAP_V3_QUOTER = '0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6'

# same addresses on all networks
addresses = {
    Network.Mainnet: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000, 100],
    },
    Network.Arbitrum: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000],
    },
    Network.Optimism: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000, 100],
    },
    Network.Base: {
        'factory': '0x33128a8fC17869897dcE68Ed026d694621f6FDfD',
        'quoter': '0x3d4e44Eb1374240CE5F1B871ab261CD16335B76a', # quoter v2
        'fee_tiers': [3000, 500, 10_000, 100],
    }
}

FEE_DENOMINATOR = 1_000_000


class UniswapV3Pool(ContractBase):
    def __init__(
        self,
        address: Address,
        token0: Address, 
        token1: Address, 
        tick_spacing: int, 
        fee: int, 
        asynchronous: bool = True
    ) -> None:
        super().__init__(address, asynchronous=asynchronous)
        self.token0 = ERC20(token0, asynchronous=asynchronous)
        self.token1 = ERC20(token1, asynchronous=asynchronous)
        self.tick_spacing = tick_spacing
        self.fee = fee

    def __contains__(self, token: Address) -> bool:
        return token in [self.token0, self.token1]
    
    def __getitem__(self, token: Address) -> ERC20:
        if token not in self:
            raise TokenNotFound(f"{token} is not in {self}")
        return ERC20(token, asynchronous=self.asynchronous)

    async def check_liquidity(self, token: AnyAddressType, block: Block) -> Optional[int]:
        if block < await contract_creation_block_async(self.address):
            return 0
        return await self[token].balance_of(self.address, block, sync=False)


class UniswapV3(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = True) -> None:
        self.asynchronous = asynchronous
        if chain.id not in addresses:
            raise UnsupportedNetwork('compound is not supported on this network')
        self.fee_tiers = addresses[chain.id]['fee_tiers']
        self.loading = False
        self.loaded = a_sync.Event()
        self._pools = {}

    def __contains__(self, asset) -> bool:
        return chain.id in addresses

    @a_sync.aka.property
    async def factory(self) -> Contract:
        return await Contract.coroutine(addresses[chain.id]['factory'])
    
    @a_sync.aka.cached_property
    async def quoter(self) -> Contract:
        quoter = addresses[chain.id]['quoter']
        try:
            return await Contract.coroutine(quoter)
        except ContractNotVerified:
            return Contract.from_abi("Quoter", quoter, UNIV3_QUOTER_ABI)

    def _encode_path(self, path) -> bytes:
        types = [type for _, type in zip(path, cycle(['address', 'uint24']))]
        return encode_abi_packed(types, path)

    def _undo_fees(self, path) -> float:
        fees = [1 - fee / FEE_DENOMINATOR for fee in path if isinstance(fee, int)]
        return math.prod(fees)
    
    @a_sync.a_sync(cache_type='memory', ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_price(
        self, 
        token: Address, 
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
        ) -> Optional[UsdPrice]:

        quoter = await self.__quoter__(sync=False)
        if block and block < await contract_creation_block_async(quoter, True):
            return None

        paths = [[token, fee, usdc.address] for fee in self.fee_tiers]
        if token != weth:
            paths += [
                [token, fee, weth.address, self.fee_tiers[0], usdc.address] for fee in self.fee_tiers
            ]

        amount_in = await ERC20(token, asynchronous=True).scale

        # TODO make this properly async after extending for brownie ContractTx
        results = await fetch_multicall(
            *[
                [quoter, 'quoteExactInput', self._encode_path(path), amount_in]
                for path in paths
            ],
            block=block,
            sync=False
        )

        outputs = [
            amount / self._undo_fees(path) / 1e6
            for amount, path in zip(results, paths)
            if amount
        ]
        return UsdPrice(max(outputs)) if outputs else None

    @a_sync.aka.cached_property
    async def pools(self) -> List[UniswapV3Pool]:
        factory = await self.__factory__(sync=False)
        pools = []
        async for logs in get_logs_asap_generator(factory.address, [factory.topics["PoolCreated"]], chronological=False):
            for event in decode_logs(logs):
                token0, token1, fee, tick_spacing, pool = event.values()
                pools.append(UniswapV3Pool(pool, token0, token1, fee, tick_spacing, asynchronous=self.asynchronous))
        return pools

    @a_sync.a_sync(ram_cache_maxsize=None)
    async def pools_for_token(self, token: Address) -> List[Address]:
        return [pool for pool in await self.__pools__(sync=False) if token in pool]

    @a_sync.a_sync(ram_cache_maxsize=10_000, ram_cache_ttl=10*60)
    async def check_liquidity(self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...] = ()) -> int:
        if block < await contract_creation_block_async(await self.__quoter__(sync=False)):
            return 0
        pools: List[UniswapV3Pool] = await self.pools_for_token(token, sync=False)
        pools = [pool for pool in pools if pool not in ignore_pools]
        return max(await asyncio.gather(*[pool.check_liquidity(token, block, sync=False) for pool in pools])) if pools else 0

try:
    uniswap_v3 = UniswapV3(asynchronous=True)
except UnsupportedNetwork:
    uniswap_v3 = None



import asyncio
import logging
from typing import Awaitable, Callable, Tuple

import a_sync

from y import convert
from y.constants import STABLECOINS
from y.datatypes import Address, AnyAddressType
from y.prices import convex, one_to_one, popsicle, solidex, yearn
from y.prices.band import band
from y.prices.chainlink import chainlink
from y.prices.dex import mooniswap
from y.prices.dex.balancer import balancer_multiplexer
from y.prices.dex.genericamm import generic_amm
from y.prices.dex.uniswap import uniswap_multiplexer
from y.prices.eth_derivs import creth, wsteth
from y.prices.gearbox import gearbox
from y.prices.lending import ib
from y.prices.lending.aave import aave
from y.prices.lending.compound import compound
from y.prices.stable_swap import (belt, ellipsis, froyo, mstablefeederpool,
                                  saddle)
from y.prices.stable_swap.curve import curve
from y.prices.synthetix import synthetix
from y.prices.tokenized_fund import basketdao, gelato, piedao, tokensets
from y.utils.logging import get_price_logger

logger = logging.getLogger(__name__)

@a_sync.a_sync(default='sync', cache_type='memory')
async def check_bucket(
    token: AnyAddressType
    ) -> str:

    import y._db.utils as db

    token_address = convert.to_address(token)
    logger = get_price_logger(token_address, block=None, extra='buckets')
    
    bucket = await db._get_token_bucket(token_address)
    if bucket:
        logger.debug('returning bucket %s from ydb', bucket)
        return bucket

    # these require neither calls to the chain nor contract initialization, just string comparisons (pretty sure)
    for bucket, check in string_matchers.items():
        if check(token):
            logger.debug("%s is %s", token_address, bucket)
            await db._set_token_bucket(token, bucket)
            return bucket
        else:
            logger.debug("%s is not %s", token_address, bucket)

    # check these first, these just require calls
    futs = [asyncio.ensure_future(_check_bucket_helper(bucket, check, token_address)) for bucket, check in calls_only.items()]
    for fut in asyncio.as_completed(futs):
        try:
            bucket, is_member = await fut
        except Exception as e:
            logger.warning("%s when checking %s. This will probably not impact your run.", e, fut)
            logger.warning(e, exc_info=True)
            continue

        if is_member:
            logger.debug("%s is %s", token_address, bucket)
            for fut in futs:
                fut.cancel()
            await db._set_token_bucket(token, bucket)
            return bucket
        else:
            logger.debug("%s is not %s", token_address, bucket)
            bucket = None

    # TODO: Refactor the below like the above
    # these require both calls and contract initializations
    if await solidex.is_solidex_deposit(token_address, sync=False):
        bucket = 'solidex'
    if await uniswap_multiplexer.is_uniswap_pool(token_address, sync=False):
        bucket = 'uni or uni-like lp'
    elif gearbox and await gearbox.is_diesel_token(token_address, sync=False):
        bucket = 'gearbox'

    # this just requires contract initialization but should go behind uniswap
    elif await aave.is_wrapped_atoken_v2(token_address, sync=False):
        bucket = 'wrapped atoken v2'
    elif await aave.is_wrapped_atoken_v3(token_address, sync=False):
        bucket = 'wrapped atoken v3'
    elif token_address in generic_amm:                                      
        bucket = 'generic amm'
    elif await mooniswap.is_mooniswap_pool(token_address, sync=False):      
        bucket = 'mooniswap lp'
    elif await compound.is_compound_market(token_address, sync=False):      
        bucket = 'compound'
    elif await _chainlink_and_band(token_address):                          
        bucket = 'chainlink and band'
    elif chainlink and await chainlink.has_feed(token_address, sync=False): 
        bucket = 'chainlink feed'
    elif synthetix and await synthetix.is_synth(token_address, sync=False): 
        bucket = 'synthetix'
    elif await yearn.is_yearn_vault(token_address, sync=False):             
        bucket = 'yearn or yearn-like'
    elif curve and await curve.get_pool(token_address, sync=False):                   
        bucket = 'curve lp'
    logger.debug("%s bucket is %s", token_address, bucket)
    if bucket:
        await db._set_token_bucket(token, bucket)
    return bucket

# these require neither calls to the chain nor contract initialization, just string comparisons (pretty sure)
string_matchers = {
    'wrapped gas coin': lambda address: address == "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
    'stable usd': lambda address: address in STABLECOINS,
    'one to one': one_to_one.is_one_to_one_token,
    'wsteth': wsteth.is_wsteth,
    'creth': creth.is_creth,
    'belt lp': belt.is_belt_lp,
    'froyo': froyo.is_froyo,
    'convex': convex.is_convex_lp,
}

# these just require calls
calls_only = {
    'atoken': aave.is_atoken,
    'balancer pool': balancer_multiplexer.is_balancer_pool,
    'ib token': ib.is_ib_token,
    'gelato': gelato.is_gelato_pool,
    'piedao lp': piedao.is_pie,
    'token set': tokensets.is_token_set,
    'ellipsis lp': ellipsis.is_eps_rewards_pool,
    'mstable feeder pool': mstablefeederpool.is_mstable_feeder_pool,
    'saddle': saddle.is_saddle_lp,
    'basketdao': basketdao.is_basketdao_index,
    'popsicle': popsicle.is_popsicle_lp,
}

async def _chainlink_and_band(token_address) -> bool:
    """ We only really need band for a short period in the beginning of fantom's history, and then we will default to chainlink once available. """
    return chainlink and await chainlink.has_feed(token_address, sync=False) and token_address in band

async def _check_bucket_helper(bucket: str, check: Callable[[Address], Awaitable[bool]], address: Address) -> Tuple[str, bool]:
    return bucket, await check(address, sync=False)




# sourcery skip: merge-assign-and-aug-assign
import asyncio
import logging
import os
from collections import defaultdict
from http import HTTPStatus
from random import randint
from time import time
from typing import Any, Callable, Dict, List, Optional

from a_sync.primitives import ThreadsafeSemaphore
from aiohttp import (BasicAuth, ClientResponse, ClientSession, ClientTimeout,
                     TCPConnector)
from aiohttp.client_exceptions import ClientError, ContentTypeError
from async_lru import alru_cache
from brownie import chain

from y import constants
from y.classes.common import UsdPrice
from y.datatypes import Address, Block
from y.networks import Network
from y.utils.dank_mids import dank_w3
from dank_mids.semaphores import BlockSemaphore

logger = logging.getLogger(__name__)

# current
header_env_names = {"X-Signer": "YPRICEAPI_SIGNER", "X-Signature": "YPRICEAPI_SIGNATURE"}
AUTH_HEADERS = {header: os.environ.get(env) for header, env in header_env_names.items()}
AUTH_HEADERS_PRESENT = all(AUTH_HEADERS.values())

# old
YPRICEAPI_USER = os.environ.get("YPRICEAPI_USER")
YPRICEAPI_PASS = os.environ.get("YPRICEAPI_PASS")
OLD_AUTH = BasicAuth(YPRICEAPI_USER, YPRICEAPI_PASS) if YPRICEAPI_USER and YPRICEAPI_PASS else None

ONE_MINUTE = 60  # some arbitrary amount of time in case the header is missing on unexpected 5xx responses
ONE_HOUR = ONE_MINUTE * 60
FALLBACK_STR = "Falling back to your node for pricing."

YPRICEAPI_TIMEOUT = ClientTimeout(int(os.environ.get("YPRICEAPI_TIMEOUT", 5 * ONE_MINUTE)))  # Five minutes is the default timeout from aiohttp.
YPRICEAPI_SEMAPHORE = BlockSemaphore(int(os.environ.get("YPRICEAPI_SEMAPHORE", 100)))

if any(AUTH_HEADERS.values()) and not AUTH_HEADERS_PRESENT:
    for header in AUTH_HEADERS:
        if not AUTH_HEADERS[header]:
            raise EnvironmentError(f'You must also pass in a value for {header_env_names[header]} in order to use ypriceAPI.')
                
should_use = not constants.SKIP_YPRICEAPI 
notified = set()
auth_notifs = defaultdict(int)
resume_at = 0
get_retry_header: Callable[[ClientResponse], int] = lambda x: int(x.headers.get("Retry-After", ONE_MINUTE))

# NOTE: if you want to bypass ypriceapi for specific tokens, have your program add the addresses to this set.
skip_tokens = set()
skip_ypriceapi = skip_tokens  # alias for backward compatability

#########################
# YPRICEAPI PUBLIC BETA #
#########################
    
_you_get = [
    "access to your desired price data more quickly...",
    "...from nodes run by yearn-affiliated big brains...",
    "...on all the networks Yearn supports."
]
_testimonials = [
    "I can now get prices for all of my useless shitcoins without waiting all day for ypricemagic to load logs.",
    "I don't need to maintain an archive node anymore and that's saving me money.", 
    "Wow, so fast!",
]
beta_announcement = "ypriceAPI is now in beta!\n\n"
beta_announcement += "Head to ypriceapi-beta.yearn.finance and sign up for access. You get:\n"
for you_get in _you_get:
    beta_announcement += f" - {you_get}\n"
beta_announcement += "\nCheck out some testimonials from our close frens:\n"
for testimonial in _testimonials:
    beta_announcement += f' - from anon{randint(0, 9999)}, "{testimonial}"\n'

def announce_beta() -> None:
    spam_your_logs_fn = logger.info if logger.isEnabledFor(logging.INFO) else print
    spam_your_logs_fn(beta_announcement)
    global should_use
    should_use = False

# TODO: Remove this when enough time has passed.
# Notify user if using old auth scheme
if OLD_AUTH is not None:
    announce_beta()
    raise NotImplementedError(
        "YPRICEAPI_USER and YPRICEAPI_PASS are no longer used.\n"
        + "Please sign up for a plan (we have a free tier) at ypriceapi-beta.yearn.finance.\n"
        + "Then, pass in the following env vars to continue using ypriceAPI:\n"
        + " - YPRICEAPI_SIGNATURE, the signature you generated on the website"
        + " - YPRICEAPI_SIGNER, the wallet you used to sign up\n"
        + "You can unset the old envs to continue using ypricemagic."
    )



class BadResponse(Exception):
    pass

@alru_cache(maxsize=1)
async def get_session() -> ClientSession:
    return ClientSession(
        os.environ.get("YPRICEAPI_URL", "https://ypriceapi-beta.yearn.finance"),
        connector=TCPConnector(verify_ssl=False),
        headers=AUTH_HEADERS,
        timeout=YPRICEAPI_TIMEOUT,
    )

@alru_cache(ttl=ONE_HOUR)
async def get_chains() -> Dict[int, str]:
    session = await get_session()
    async with session.get("/chains") as response:
        chains = await read_response(response) or {}
    return {int(k): v for k, v in chains.items()}

@alru_cache(ttl=ONE_HOUR)
async def chain_supported(chainid: int) -> bool:
    if chainid in await get_chains():
        return True
    logger.info('ypriceAPI does not support %s at this time.', Network.name())
    return False

async def get_price(
    token: Address,
    block: Optional[Block]
) -> Optional[UsdPrice]:
    
    if not AUTH_HEADERS_PRESENT:
        announce_beta()
        return None
        
    if time() < resume_at:
        # NOTE: The reason we are here has already been logged.
        return None

    if block is None:
        block = await dank_w3.eth.block_number

    async with YPRICEAPI_SEMAPHORE[block]:
        try:
            if not await chain_supported(chain.id):
                return None
            session = await get_session()
            async with session.get(f'/get_price/{chain.id}/{token}?block={block}') as response:
                if price := await read_response(response, token, block):
                    return UsdPrice(price)
        except asyncio.TimeoutError:
            logger.warning(f'ypriceAPI timed out for {token} at {block}.{FALLBACK_STR}')
        except ContentTypeError:
            raise
        except ClientError as e:
            logger.warning(f'ypriceAPI {e.__class__.__name__} for {token} at {block}.{FALLBACK_STR}')

async def read_response(response: ClientResponse, token: Optional[Address] = None, block: Optional[Block] = None) -> Optional[Any]:
    
    # 200
    if response.status == HTTPStatus.OK:
        try:
            return await response.json()
        except ContentTypeError as e:
            msg = await response.json(content_type=None)
            if msg:
                raise BadResponse(msg) from e
            logger.warning(f'ypriceAPI returned status code {_get_err_reason(response)} with response None. Must investigate.')

    # 401
    elif response.status == HTTPStatus.UNAUTHORIZED:
        if HTTPStatus.UNAUTHORIZED not in notified:
            logger.error(f'Your provided ypriceAPI credentials are not authorized for use.{FALLBACK_STR}')
            notified.add(HTTPStatus.UNAUTHORIZED)
            
    # 404
    elif response.status == HTTPStatus.NOT_FOUND and token and block:
        logger.debug("Failed to get price from API: %s at %s", token, block)

    # Server Errors
    
    # 502 & 503
    elif response.status in {HTTPStatus.BAD_GATEWAY, HTTPStatus.SERVICE_UNAVAILABLE}:
        logger.warning("ypriceAPI returned status code %s", _get_err_reason(response))
        try:
            msg = await response.json(content_type=None) or await response.text()
        except Exception:
            logger.warning('exception decoding ypriceapi %s response.%s', response.status, FALLBACK_STR, exc_info=True)
            msg = ''
        if msg:
            logger.warning(msg)
        _set_resume_at(get_retry_header(response))

    else:
        msg = f'ypriceAPI returned status code {_get_err_reason(response)}'
        if token and block:
            msg += f' for {token} at {block}.{FALLBACK_STR}'
        logger.warning(msg)
            

def _get_err_reason(response: ClientResponse) -> str:
    if response.reason is None:
        return f"[{response.status}]"
    ascii_encodable_reason = response.reason.encode(
        "ascii", "backslashreplace"
    ).decode("ascii")
    return f"[{response.status} {ascii_encodable_reason}]"
    
def _set_resume_at(retry_after: float) -> None:
    global resume_at
    logger.info("Falling back to your node for %s minutes.", int(retry_after/60))
    resume_from_this_err_at = time() + retry_after
    if resume_from_this_err_at > resume_at:
        resume_at = resume_from_this_err_at
